/**********************************************************************
 *                   GNU General Public License                       *
 * This file is a part of HIM.                                        *
 *                                                                    *
 * HIM is free software; you can redistribute it and/or modify it and *
 * are expected to follow the terms of the GNU General Public License *
 * as published by the Free Software Foundation; either version 2 of  *
 * the License, or (at your option) any later version.                *
 *                                                                    *
 * HIM is distributed in the hope that it will be useful, but WITHOUT *
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public   *
 * License for more details.                                          *
 *                                                                    *
 * For the full text of the GNU General Public License,               *
 * write to: Free Software Foundation, Inc.,                          *
 *           675 Mass Ave, Cambridge, MA 02139, USA.                  *
 * or see:   http://www.gnu.org/licenses/gpl.html                     *
 **********************************************************************/

/********+*********+*********+*********+*********+*********+*********+*
 *                                                                    *
 *  By Robert Hallberg, 2002                                          *
 *                                                                    *
 *    This file contains an example of the code that is needed to set *
 *  up and use a set (in this case one) of dynamically passive tracers*
 *                                                                    *
 *    A single subroutine is called from within each file to register *
 *  each of the tracers for reinitialization and advection and to     *
 *  register the subroutine that initializes the tracers and set up   *
 *  their output and the subroutine that does any tracer physics or   *
 *  chemistry along with diapycnal mixing (included here because some *
 *  tracers may float or swim vertically or dye diapycnal processes). *
 *                                                                    *
 *                                                                    *
 *  Variables written all in capital letters are defined in init.h    *
 *                                                                    *
 *     A small fragment of the grid is shown below:                   *
 *                                                                    *
 *    j+1  x ^ x ^ x   At x:  q                                       *
 *    j+1  > o > o >   At ^:  v                                       *
 *    j    x ^ x ^ x   At >:  u                                       *
 *    j    > o > o >   At o:  h, tr                                   *
 *    j-1  x ^ x ^ x                                                  *
 *        i-1  i  i+1  At x & ^:                                      *
 *           i  i+1    At > & o:                                      *
 *                                                                    *
 *  The boundaries always run through q grid points (x).              *
 *                                                                    *
 ********+*********+*********+*********+*********+*********+*********+*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <init.h>
#include <HIM_io.h>
#include <HIM.h>
#include <metrics.h>

#define NTR 11               /* The number of tracers in this module. */
extern struct params HIM_params;   /* HIM_params is a structure that  */
                                   /* contains a number of parameters */
                                   /* of the simulation which can be  */
                                   /* modified from a namelist file   */
                                   /* at run time.                    */

static double tr[NTR][NZ][NYMEM][NXMEM];/* Tracer concentration g m-3.*/

extern double D[NYMEM][NXMEM];       /* Basin depth, in m.            */
extern double h[3][NZ][NYMEM][NXMEM];/* Layer thickness, in m.        */

extern double geolath[NYMEM][NXMEM]; /* The geographic latitude and   */
extern double geolonh[NYMEM][NXMEM]; /* longitude at h points in      */
                                     /* degrees or m.                 */

extern char directory2[];            /* The directory to use to save  */
                                     /* output files.                 */

extern int pe_here;                  /* The current processor's label.*/
extern int X1off, Y1off;             /* The offset in the global      */
                                     /* indices of X1 and Y1 on the   */
                                     /* current processor, relative   */
                                     /* to X1 and Y1 on PE 0.         */
extern int nx, ny;                   /* The largest index value in the*/
                                     /* x- and y- directions of the   */
                                     /* local computational domain.   */

void USER_register_tracer_example(void);

static void USER_initialize_tracer(int restart, time_type day);
static void tracer_column_physics(double h_old[NZ][NYMEM][NXMEM],
         double h_new[NZ][NYMEM][NXMEM], double ea[NZ][NYMEM][NXMEM],
         double eb[NZ][NYMEM][NXMEM], struct forcing fluxes, double dt);

void USER_register_tracer_example(void) {
/* This subroutine is used to register tracer fields and subroutines  */
/* to be used with HIM.                                               */
  int m;

  /* Register the tracer for the restart file.                        */
  for (m=0;m<NTR;m++) {
    vardesc vd = {"tr","Tracer",'h','L','s',"kg kg-1", 'd'};
    sprintf(vd.name,"tr%1d",m);
    sprintf(vd.longname,"Concentration of Tracer %1d",m);
    register_restart_field(tr[m][0][0], NULL, vd, 1);
  }
  /* Register the tracer for horizontal advection & diffusion.        */
  for (m=0;m<NTR;m++) {
    char name[32];
    sprintf(name,"tr%1d",m);
    register_tracer(&tr[m],NZ,name);
  }

  register_tracer_init_fn(&(USER_initialize_tracer));

  register_tracer_column_fn(&(tracer_column_physics));
}

static void USER_initialize_tracer(int restart, time_type day) {
/*   This subroutine initializes the NTR tracer fields in tr[][][][]  */
/* and it sets up the tracer output.                                  */

/* Arguments: restart - 1 if the fields have already been read from   */
/*                     a restart file.                                */
/*  (in)      day - Time of the start of the run.                     */
  int i, j, k, m;
  if (!restart) {
    int from_file = 0;
    if (from_file!=0) {
/*  Read the tracer concentrations from a netcdf file.                */
      char name[16], filename[200];
      int err, cdfid;

      strcpy(filename,HIM_params.inputdir);
      strcat(filename,"/MESO_IC_TR_2.cdf");

      err = open_input_file(filename, &cdfid, NULL);
      if (err != 0) {printf("Unable to open %s.\n",filename); quit(-10); }

      for (m=0;m<NTR;m++) {
/* This line should put "trX" into name, where X = m.                 */
/* If multiple tracers with names other than tr1, tr2, tr3, ... are   */
/* used, these will have to be specified somehow.                     */
        sprintf(name,"tr%-1d",m+1);
        Read_Field(cdfid, name, 'h', NZ, 0, 1, &tr[m][0][0][0]);
      }

      close_file(&cdfid);
    } else {
      for (m=0;m<NTR;m++)
        for (k=0;k<=NZ-1;k++) for (j=Y1;j<=ny;j++) for (i=X1;i<=nx;i++)
          tr[m][k][j][i] = EPSILON * EPSILON;

/*    This sets a stripe of tracer across the basin.                  */
      for (m=1;m<NTR ;m++) {
        for (j=Y1;j<=ny;j++) {
          double tr_y;/*Initial zonally uniform tracer concentrations.*/
          if (m<=5)
            tr_y = ((geolath[j][i] > (300.0+50.0*(double)m)) &&
                    (geolath[j][i] < (350.0+50.0*(double)m))) ? 1.0 : 0.0;
          else tr_y = 0.0;

          for (i=X1;i<=nx;i++) for (k=0;k<=NZ-1;k++) {
/*    This adds the stripes of tracer to every layer.                 */
            tr[m][k][j][i] += tr_y;
          }
        }
      }

      if (NTR>6) for (j=Y1;j<=ny;j++) {
        for (i=X1;i<=nx;i++) {
          double e[NZ+1], e_top, e_bot, d_tr;
          e[NZ] = -D[j][i];
          for (k=NZ-1;k>=0;k--) {
            e[k] = e[k+1] + h[0][k][j][i];
            for (m=6;m<NTR;m++) {
              e_top = -600.0*m + 3000.0;
              e_bot = -600.0*m + 2700.0;
              if (e_top < e[k]) {
                if (e_top < e[k+1]) d_tr = 0.0;
                else if (e_bot < e[k+1]) d_tr = (e_top - e[k+1]) / h[0][k][j][i];
                else d_tr = (e_top - e_bot) / h[0][k][j][i];
              }
              else if (e_bot < e[k]) {
                if (e_bot < e[k+1]) d_tr = 1.0;
                else d_tr = (e[k] - e_bot) / h[0][k][j][i];
              }
              else d_tr = 0.0;

              if (h[0][k][j][i] < 2.0*EPSILON) d_tr = 0.0;

              tr[m][k][j][i] += d_tr;
            }
          }
        }
      }
    }
  }

#ifdef SPONGE
/*   If sponges are used, this example damps tracers in sponges in the*/
/* northern half of the domain to 1 and tracers in the southern half  */
/* to 0.  For any tracers that are not damped in the sponge, the call */
/* to set_up_sponge_field can simply be omitted.                      */
  {
    double temp[NZ][NYMEM][NXMEM];
    int i, j, k;
    for (i=X1;i<=nx;i++) for (j=Y1;j<=ny;j++) for (k=0;k<NZ;k++) {
      temp[k][j][i] = ((geolath[j][i] > 700.0) && (k>NZ/2)) ? 1.0 : 0.0;
    }
    for (m=0;m<1;m++) set_up_sponge_field(temp,&(tr[m]),NZ);
  }
#endif

/*   Set up the tracer output.  See initialize_output.c for a fuller  */
/* example of how the output is specified.                            */
/* Uncomment the following line to put tracers into their own file.   */
/* Otherwise, they go into the same file as the previous variable.    */
   specify_saves_file("tracer", directory2, 's', day, HIM_params.daysave,
       HIM_params.daymax, t_mult(HIM_params.saveint, 4), HIM_params.saves_per_file);

  /* respecify_saves_file("save", directory2); */

  for (m=0;m<NTR;m++) {
    vardesc vd = {"tr","Tracer",'h','L','s',"kg kg-1", 'f'};
    sprintf(vd.name,"tr%1d",m);
    sprintf(vd.longname,"Concentration of Tracer %1d",m);
    set_up_save_field(&tr[m][0],NULL,vd);
  }
}

static void tracer_column_physics(double h_old[NZ][NYMEM][NXMEM],
    double h_new[NZ][NYMEM][NXMEM], double ea[NZ][NYMEM][NXMEM],
    double eb[NZ][NYMEM][NXMEM], struct forcing fluxes, double dt) {
/*   This subroutine applies diapycnal diffusion and any other column */
/* tracer physics or chemistry to the tracers from this file.         */
/* This is a simple example of a set of advected passive tracers.     */

/* Arguments: h_old -  Layer thickness before entrainment, in m.      */
/*  (in)      h_new -  Layer thickness after entrainment, in m.       */
/*  (in)      ea - an array to which the amount of fluid entrained    */
/*                 from the layer above during this call will be      */
/*                 added, in m.                                       */
/*  (in)      eb - an array to which the amount of fluid entrained    */
/*                 from the layer below during this call will be      */
/*                 added, in m.                                       */
/*  (in)      fluxes - A structure containing pointers to any possible*/
/*                     forcing fields.  Unused fields have NULL ptrs. */
/*  (in)      dt - The amount of time covered by this call, in s.     */
/*                                                                    */
/* The arguments to this subroutine are redundant in that             */
/*     h_new[k] = h_old[k] + ea[k] - eb[k-1] + eb[k] - ea[k+1]        */

  double hold0[NXMEM];      /* The original topmost layer thickness,  */
                            /* with surface mass fluxes added back, m.*/
  double b1[NXMEM];         /* b1 and c1 are variables used by the    */
  double c1[NZ][NXMEM];     /* tridiagonal solver.                    */
  int i, j, k, m;
/* The following array (trdc) determines the behavior of the tracer   */
/* diapycnal advection.  The first element is 1 if tracers are        */
/* passively advected.  The second and third are the concentrations   */
/* to which downwelling and upwelling water are set, respectively.    */
/* For most (normal) tracers, the appropriate vales are {1,0,0}.      */

/*   Uncomment the following line to dye both upwelling and downwelling.*/
  /* double trdc[3] = {0.0,1.0,1.0}; */
/*   Uncomment the following line to dye downwelling.                 */
  /* double trdc[3] = {0.0,1.0,0.0}; */
  /*   Uncomment the following line to dye upwelling.                 */
  /* double trdc[3] = {0.0,0.0,1.0}; */
/*   Uncomment the following line for tracer concentrations to be set */
/* to zero in any diapycnal motions.                                  */
  /* double trdc[3] = {0.0,0.0,0.0}; */
/*   Uncomment the following line for most "physical" tracers, which  */
/* are advected diapycnally in the usual manner.                      */
  double trdc[3] = {1.0,0.0,0.0};

  for (j=Y1;j<=ny;j++) {
    for (i=X1;i<=nx;i++) {
/*   The following line is appropriate for quantities like salinity   */
/* that are left behind by evaporation, and any surface fluxes would  */
/* be explicitly included in the flux structure.                      */
      hold0[i] = h_old[0][j][i] - ea[0][j][i];
/*   The following line is appropriate for quantities like temperature*/
/* that can be assumed to have the same concentration in evaporation  */
/* as they had in the water.  The explicit surface fluxes here would  */
/* reflect differences in concentration from the ambient water, not   */
/* the absolute fluxes.                                               */
  /*  hold0[i] = h_old[0][j][i];  */

      b1[i]=1.0/(h_old[0][j][i]+eb[0][j][i]);
      for (m=0;m<NTR;m++)
        tr[m][0][j][i] = b1[i]*(hold0[i]*tr[m][0][j][i] +
                                 trdc[2]*eb[0][j][i]);
/*      Add any surface tracer fluxes to the preceding line.          */
    }
    for (k=1;k<=NZ-1;k++) {
      for (i=X1;i<=nx;i++) {
        c1[k][i] = trdc[0] * eb[k-1][j][i] * b1[i];
        b1[i]=1.0/(h_old[k][j][i] + eb[k][j][i] +
                   (1.0-c1[k][i])*ea[k][j][i]);
        for (m=0;m<NTR;m++)
          tr[m][k][j][i] = b1[i] * (h_old[k][j][i]*tr[m][k][j][i] +
                   ea[k][j][i]*(trdc[0]*tr[m][k-1][j][i]+trdc[1]) +
                   eb[k][j][i]*trdc[2]);
      }
    }
    for (k=NZ-2;k>=0;k--) {
      for (i=X1;i<=nx;i++) {
        for (m=0;m<NTR;m++)
          tr[m][k][j][i] += c1[k+1][i]*tr[m][k+1][j][i];
      }
    }
  }

}
