/**********************************************************************
 *                   GNU General Public License                       *
 * This file is a part of HIM.                                        *
 *                                                                    *
 * HIM is free software; you can redistribute it and/or modify it and *
 * are expected to follow the terms of the GNU General Public License *
 * as published by the Free Software Foundation; either version 2 of  *
 * the License, or (at your option) any later version.                *
 *                                                                    *
 * HIM is distributed in the hope that it will be useful, but WITHOUT *
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public   *
 * License for more details.                                          *
 *                                                                    *
 * For the full text of the GNU General Public License,               *
 * write to: Free Software Foundation, Inc.,                          *
 *           675 Mass Ave, Cambridge, MA 02139, USA.                  *
 * or see:   http://www.gnu.org/licenses/gpl.html                     *
 **********************************************************************/

/********+*********+*********+*********+*********+*********+*********+*
 *                                                                    *
 *  By Robert Hallberg, April 1994 - June 2002                        *
 *                                                                    *
 *    This program contains the subroutines that time steps the       *
 *  linearized barotropic equations.  btstep is used to actually      *
 *  time step the barotropic equations.  btcalc determines the        *
 *  vertical averages.  barotropic_init is called once at the start   *
 *  of the simulation, and initializes the constants which are used   *
 *  in btstep.                                                        *
 *                                                                    *
 *    btstep uses a forward-backward based scheme to time step the    *
 *  barotropic equations.  The average barotropic acceleration is     *
 *  calculated and used to step the layer velocities. In addition,    *
 *  the average barotropic velocity is taken over all the time steps. *
 *  This average velocity is used to step the baroclinic continuity   *
 *  equations.                                                        *
 *    btstep takes 7 arguments:  the time level of the initial        *
 *  fields (lin), the time level of the final fields (lout), the      *
 *  amount of time over which to integrate (bt), zonal and meridional *
 *  barotropic forcing fields (ubt_force and vbt_force), and the      *
 *  difference between these forcing fields and the realized          *
 *  accelerations from the barotropic calculations (u_accel_bt and    *
 *  v_accel_bt). lin and lout may not be the same.                    *
 *                                                                    *
 *    The spatial discretization of the continuity equation is second *
 *  order accurate.  A flux conservative form is used to guarantee    *
 *  global conservation of volume.  The spatial discretization of the *
 *  momentum equation is second order accurate.  The Coriolis force   *
 *  is written in a form which does not contribute to the energy      *
 *  tendency and which conserves linearized potential vorticity, f/D. *
 *  These terms are exactly removed from the baroclinic momentum      *
 *  equations, so the linearization of vorticity advection will not   *
 *  degrade the overall solution.                                     *
 *                                                                    *
 *    btcalc takes 5 arguments, the two layer velocity components,    *
 *  the thicknesses, the time level of the barotropic fields, and a   *
 *  flag to indicate whether the barotropic velocities and the        *
 *  fractional thicknesses are to be calculated.  (Otherwise just a   *
 *  correction to the free surface height is calculated.) btcalc      *
 *  interpolates the thicknesses onto velocity grid points using the  *
 *  second order accurate scheme                                      *
 *    h(at velocity points) = 2.0 * h+ * h- / (h+ + h-)               *
 *  where h+ and h- are the thicknesses adjacent to the velocity grid *
 *  point.  btcalc then calculates the fraction of the total depth in *
 *  each layer at velocity grid points (frhatu & frhatv).  btcalc     *
 *  calculates the depth averaged (barotropic) velocities.  These     *
 *  first two fields are only calculated if so indicated.  The        *
 *  difference between the free surface height from the layer         *
 *  thicknesses and the free surface height from the barotropic       *
 *  integration is always calculated.                                 *
 *                                                                    *
 *    barotropic_const_setup takes no arguments.                      *
 *                                                                    *
 *  Variables written all in capital letters are defined in init.h    *
 *                                                                    *
 *     A small fragment of the grid is shown below:                   *
 *                                                                    *
 *    j+1  x ^ x ^ x   At x:  q, f                                    *
 *    j+1  > o > o >   At ^:  vbt, vbt_force, v_accel_bt, vbtav       *
 *    j    x ^ x ^ x   At >:  ubt, ubt_force, u_accel_bt, ubtav, amer *
 *    j    > o > o >   At o:  eta, D                                  *
 *    j-1  x ^ x ^ x                                                  *
 *        i-1  i  i+1                                                 *
 *           i  i+1                                                   *
 *                                                                    *
 *  The boundaries always run through q grid points (x).              *
 *                                                                    *
 ********+*********+*********+*********+*********+*********+*********+*/

#include <math.h>

#include <init.h>
#include <HIM.h>
#include <metrics.h>
#include <HIM_io.h>

extern double D[NYMEM][NXMEM];       /* Basin depth, in m.            */
extern double f[NYMEM][NXMEM];       /* The Coriolis parameter in s-1.*/

extern double umask[NYMEM][NXMEM];   /* _mask are 1 over ocean and 0  */
extern double vmask[NYMEM][NXMEM];   /* over land on the u & v grids. */

extern struct params HIM_params;   /* HIM_params is a structure that  */
                                   /* contains a number of parameters */
                                   /* of the simulation which can be  */
                                   /* modified from a namelist file at*/
                                   /* run time.                       */

extern struct diag_fld diag;       /* A structure containing pointers */
                                   /* to diagnostic fields which might*/
                                   /* be calculated.                  */

#ifdef SPLIT

static double ubt[3][NYMEM][NXMEM];  /* ubt & vbt are the barotropic  */
static double vbt[3][NYMEM][NXMEM];  /* zonal & meridional velocities,*/
                                     /* in m s-1.                     */
static double eta[3][NYMEM][NXMEM];  /* The free surface height in m. */

static double ubtav[NYMEM][NXMEM];   /* ubtav & vbtav are the baro-   */
static double vbtav[NYMEM][NXMEM];   /* tropic zonal and meridional   */
                                     /* velocities averaged over the  */
                                     /* baroclinic time step, m s-1.  */

static struct gtot gtot[NYMEM][NXMEM]; /* gtot is the effective total */
                                     /* reduced gravity used to       */
                                     /* relate free surface height    */
                                     /* deviations to pressure forces */
                                     /* in the barotropic momentum    */
                                     /* equations, including GFS and  */
                                     /* baroclinic contributions and  */
                                     /* in m s-2.                     */

static double frhatu[NZ][NYMEM][NXMEM];/* Fraction of the total column*/
static double frhatv[NZ][NYMEM][NXMEM];/* thickness interpolated to u */
                                     /* or v grid points in each      */
                                     /* layer, nondimensional.        */

static double eta_cor[NYMEM][NXMEM]; /* The difference between the    */
                                     /* free surface height from the  */
                                     /* barotropic calculation and    */
                                     /* the sum of the layer thick-   */
                                     /* nesses.  This difference is   */
                                     /* imposed as a forcing term in  */
                                     /* the barotropic calculation    */
                                     /* over a baroclinic timestep,   */
                                     /* in units of m.                */
# ifdef BOUND_BT_CORRECTION
static double eta_cor_bound[NYMEM][NXMEM];/* A limit on the rate at   */
                                     /* which eta_cor can be applied  */
                                     /* while avoiding instability,   */
                                     /* in units of m s-1.            */
# endif

static double azon[NYMEM][NXMEM];    /* _zon & _mer are the values    */
static double bzon[NYMEM][NXMEM];    /* of the Coriolis force which   */
static double czon[NYMEM][NXMEM];    /* are applied to the neighboring*/
static double dzon[NYMEM][NXMEM];    /* values of vbtav & ubtav,      */
static double amer[NYMEM][NXMEM];    /* respectively to get the       */
static double bmer[NYMEM][NXMEM];    /* barotropic inertial rotation, */
static double cmer[NYMEM][NXMEM];    /* in units of s-1.              */
static double dmer[NYMEM][NXMEM];

static double Datu[NYMEM][NXMEM];    /* Basin depth at u-velocity     */
                                     /* grid points times the y-grid  */
                                     /* spacing, in m2.               */
static double Datv[NYMEM][NXMEM];    /* Basin depth at v-velocity     */
                                     /* grid points times the x-grid  */
                                     /* spacing, in m2.               */
static double IDatu[NYMEM][NXMEM];   /* Inverse of the basin depth at */
static double IDatv[NYMEM][NXMEM];   /* u or v points, in m-1.        */

extern int nx, ny;                   /* The largest index value in the*/
                                     /* x- and y- directions of the   */
                                     /* local computational domain.   */
extern double bt_body_force[NYMEM][NXMEM]; /* The pattern of the body force. */
extern double bt_body_force_stren; /* The strength of the body force, */
                                   /* in units of m s-2.              */

void btstep(int lin, int lout, double dt, double ubt_force[NYMEM][NXMEM],
            double vbt_force[NYMEM][NXMEM], double u_accel_bt[NYMEM][NXMEM],
            double v_accel_bt[NYMEM][NXMEM]) {
/*    This subroutine time steps the barotropic equations explicitly. */
/*  For gravity waves, anything between a forward-backward scheme     */
/*  and a simulated backward Euler scheme is used, with bebt between  */
/*  0.0 and 1.0 determining the scheme.  In practice, bebt must be of */
/*  order 0.2 or greater.  A forward-backward treatment of the        */
/*  Coriolis terms is always used.                                    */

/* Arguments: lin - The time level of the starting barotropic fields. */
/*  (in)      lout - The time level of the final barotropic fields.   */
/*  (in)      dt - The time increment to integrate over.              */
/*  (in)      ubt_force - The zonal barotropic forcing, in m s-2.     */
/*  (in)      vbt_force - The meridional barotropic forcing, in m s-2.*/
/*  (out)     u_accel_bt - The difference between the accelerations   */
/*  (out)     v_accel_bt - from the barotropic calculation and the    */
/*                         forcing terms, in m s-2.                   */

  double u_RHS[NYMEM][NXMEM]; /* u_RHS and v_RHS are the vertical     */
  double v_RHS[NYMEM][NXMEM]; /* averages of all of the accelerations */
                              /* that are not explicitly included in  */
                              /* the barotropic equations, in m s-2.  */
  double dtbt, bebt;
  double Instep, Idt;
  int nstep, do_ave;
  int i, j, n, lpr;
  static int first_call = 1;
  static int av_ref_PFu_bt, av_ref_PFv_bt, av_ref_Coru_bt, av_ref_Corv_bt;
  static int av_ref_Nonlnu_bt, av_ref_Nonlnv_bt, av_ref_ubt_flux, av_ref_vbt_flux;

  nstep = (int) ceil(dt/HIM_params.dtbt - 0.0001);
  Instep = 1.0 / (double) nstep;
  dtbt = dt * Instep;
  bebt = HIM_params.bebt;
  Idt = 1.0 / dt;

  if (first_call == 1) {
    av_ref_PFu_bt = get_field_ref((double (*)[NYMEM][NXMEM]) diag.PFu_bt);
    av_ref_PFv_bt = get_field_ref((double (*)[NYMEM][NXMEM]) diag.PFv_bt);
    av_ref_Coru_bt = get_field_ref((double (*)[NYMEM][NXMEM]) diag.Coru_bt);
    av_ref_Corv_bt = get_field_ref((double (*)[NYMEM][NXMEM]) diag.Corv_bt);
    av_ref_Nonlnu_bt = get_field_ref((double (*)[NYMEM][NXMEM]) diag.Nonlnu_bt);
    av_ref_Nonlnv_bt = get_field_ref((double (*)[NYMEM][NXMEM]) diag.Nonlnv_bt);
    av_ref_ubt_flux = get_field_ref((double (*)[NYMEM][NXMEM]) diag.ubt_flux);
    av_ref_vbt_flux = get_field_ref((double (*)[NYMEM][NXMEM]) diag.vbt_flux);

    first_call = 0;
  }
  do_ave = query_averaging_enabled();

/*    Here the vertical average accelerations due to the Coriolis,    */
/*  advective, pressure gradient and horizontal viscous terms in the  */
/*  layer momentum equations are calculated.  These will be used to   */
/*  determine the difference between the accelerations due to the     */
/*  average of the layer equations and the barotropic calculation.    */

  for (j=Y1;j<=ny;j++) {
    for (i=X1;i<=nx;i++) {
      u_RHS[j][i] = ubt_force[j][i] -
          ((azon[j][i] * vbtav[j][i+1] + czon[j][i] * vbtav[j-1][i]) +
           (bzon[j][i] * vbtav[j][i] + dzon[j][i] * vbtav[j-1][i+1]))
          + bt_body_force_stren * bt_body_force[j][i];
      v_RHS[j][i] = vbt_force[j][i] +
          ((amer[j][i-1] * ubtav[j][i-1] + cmer[j+1][i] * ubtav[j+1][i]) +
           (bmer[j][i] * ubtav[j][i] + dmer[j+1][i-1] * ubtav[j+1][i-1]));
# ifdef BOUND_BT_CORRECTION
      if (fabs(eta_cor[j][i]) > dt*eta_cor_bound[j][i]) {
        eta_cor[j][i] = dt*((eta_cor[j][i] > 0) ?
                        eta_cor_bound[j][i] : -eta_cor_bound[j][i]);
      }
# endif
    }
  }
  for (j=Y1-1;j<=ny;j++) {
    for (i=X1-1;i<=nx;i++) {
      ubtav[j][i] = 0.0;
      vbtav[j][i] = 0.0;
      if (diag.PFu_bt != NULL) diag.PFu_bt[j][i] = 0.0;
      if (diag.PFv_bt != NULL) diag.PFv_bt[j][i] = 0.0;
      if (diag.Coru_bt != NULL) diag.Coru_bt[j][i] = 0.0;
      if (diag.Corv_bt != NULL) diag.Corv_bt[j][i] = 0.0;
    }
  }

  pass_var_2d(ubt[lin], 1, 1, 1, 0.0, 1, 0, TO_SOUTH+TO_EAST,U_PT+VECTOR);
  pass_var_2d(vbt[lin], 1, 1, 1, 0.0, 0, 0, TO_NORTH+TO_WEST,V_PT+VECTOR);

/*  The following loop contains all of the time steps.                */
  lpr = lin;

  for (n=1;n<=nstep;n++) {

    for (j=Y1;j<=ny;j++) {
      for (i=X1;i<=nx;i++) {
        eta[2][j][i] = eta[lpr][j][i] + Instep*eta_cor[j][i] + dtbt * IDXDYh(j,i) *
            ((Datu[j][i-1]*ubt[lpr][j][i-1] - Datu[j][i]*ubt[lpr][j][i]) +
             (vbt[lpr][j-1][i]*Datv[j-1][i] - vbt[lpr][j][i]*Datv[j][i]));
        ubtav[j][i] += (1.0-bebt)*ubt[lpr][j][i];
        vbtav[j][i] += (1.0-bebt)*vbt[lpr][j][i];
      }
    }
    pass_var_2d(eta[2], 1, 1, 1, 0.0, 1, 0, TO_SOUTH+TO_WEST,0);

    if (n%2==1) {
      for (j=Y1;j<=ny;j++) {
        for (i=X1;i<=nx;i++) {
          double Cor, gradP;
          Cor = -((amer[j][i-1] * ubt[lpr][j][i-1] +
                  cmer[j+1][i] * ubt[lpr][j+1][i]) +
                 (bmer[j][i] * ubt[lpr][j][i] +
                  dmer[j+1][i-1] * ubt[lpr][j+1][i-1]));
          gradP = (eta[2][j][i]*gtot[j][i].N -
                   eta[2][j+1][i]*gtot[j+1][i].S) * IDYv(j,i);
          vbt[lout][j][i] = vmask[j][i] * (vbt[lpr][j][i]
                          + dtbt * (v_RHS[j][i] + Cor + gradP));
          if (diag.PFv_bt != NULL) diag.PFv_bt[j][i] += gradP;
          if (diag.Corv_bt != NULL) diag.Corv_bt[j][i] += Cor;
        }
      }
      pass_var_2d(vbt[lout], 1, 1, 1, 0.0, 1, 0, TO_NORTH+TO_WEST,V_PT+VECTOR);
    }

    for (j=Y1;j<=ny;j++) {
      for (i=X1;i<=nx;i++) {
        double Cor, gradP;
        Cor = (azon[j][i] * vbt[lout][j][i+1] +
               czon[j][i] * vbt[lout][j-1][i]) +
              (bzon[j][i] * vbt[lout][j][i] +
               dzon[j][i] * vbt[lout][j-1][i+1]);
        gradP = (eta[2][j][i]*gtot[j][i].E -
                 eta[2][j][i+1]*gtot[j][i+1].W) * IDXu(j,i);
        ubt[lout][j][i] = umask[j][i] * (ubt[lpr][j][i]
                        + dtbt * (u_RHS[j][i] + Cor + gradP));
        if (diag.PFu_bt != NULL) diag.PFu_bt[j][i] += gradP;
        if (diag.Coru_bt != NULL) diag.Coru_bt[j][i] += Cor;
      }
    }
    pass_var_2d(ubt[lout], 1, 1, 1, 0.0, 1, 0, TO_SOUTH+TO_EAST,U_PT+VECTOR);

    if (n%2!=1) {
      for (j=Y1;j<=ny;j++) {
        for (i=X1;i<=nx;i++) {
          double Cor, gradP;
          Cor = -((amer[j][i-1] * ubt[lout][j][i-1] +
                   bmer[j][i] * ubt[lout][j][i]) +
                  (cmer[j+1][i] * ubt[lout][j+1][i] +
                   dmer[j+1][i-1] * ubt[lout][j+1][i-1]));
          gradP = (eta[2][j][i]*gtot[j][i].N -
                   eta[2][j+1][i]*gtot[j+1][i].S) * IDYv(j,i);
          vbt[lout][j][i] = vmask[j][i] * (vbt[lpr][j][i]
                          + dtbt * (v_RHS[j][i] + Cor + gradP));
          if (diag.PFv_bt != NULL) diag.PFv_bt[j][i] += gradP;
          if (diag.Corv_bt != NULL) diag.Corv_bt[j][i] -= Cor;
        }
      }
      pass_var_2d(vbt[lout], 1, 1, 1, 0.0, 1, 0, TO_NORTH+TO_WEST,V_PT+VECTOR);
    }

    for (j=Y1;j<=ny;j++) {
      for (i=X1;i<=nx;i++) {
        eta[lout][j][i] = (1.0-bebt)*eta[2][j][i] + bebt*Instep*eta_cor[j][i] +
             bebt*(eta[lpr][j][i] + dtbt * IDXDYh(j,i) *
            ((Datu[j][i-1]*ubt[lout][j][i-1] - Datu[j][i]*ubt[lout][j][i]) +
             (vbt[lout][j-1][i]*Datv[j-1][i] - vbt[lout][j][i]*Datv[j][i])));
        ubtav[j][i] += bebt*ubt[lout][j][i];
        vbtav[j][i] += bebt*vbt[lout][j][i];
      }
    }


    lpr = lout;
  }

  for (j=Y1;j<=ny;j++) {
    for (i=X1;i<=nx;i++) {
      ubtav[j][i] *= Instep;
      vbtav[j][i] *= Instep;

      eta[2][j][i] = 0.5*(eta[0][j][i] + eta[1][j][i]);
      u_accel_bt[j][i] = (ubt[lout][j][i] - ubt[lin][j][i])*Idt - ubt_force[j][i];
      v_accel_bt[j][i] = (vbt[lout][j][i] - vbt[lin][j][i])*Idt - vbt_force[j][i];

      if (diag.PFu_bt != NULL) diag.PFu_bt[j][i] *= Instep;
      if (diag.PFv_bt != NULL) diag.PFv_bt[j][i] *= Instep;
      if (diag.Coru_bt != NULL) diag.Coru_bt[j][i] *= Instep;
      if (diag.Corv_bt != NULL) diag.Corv_bt[j][i] *= Instep;
      if (diag.Nonlnu_bt != NULL) diag.Nonlnu_bt[j][i] = u_RHS[j][i];
      if (diag.Nonlnv_bt != NULL) diag.Nonlnv_bt[j][i] = v_RHS[j][i];
    }
  }
  pass_var_2d(eta[2], 1, 1, 1, 0.0, 1, 0, TO_SOUTH+TO_WEST,0);
  pass_var_2d(vbtav, 1, 1, 1, 0.0, 0, 0, TO_NORTH+TO_WEST,V_PT+VECTOR);
  pass_var_2d(ubtav, 1, 1, 1, 0.0, 0, 0, TO_SOUTH+TO_EAST,U_PT+VECTOR);

  if (diag.ubt_flux != NULL) for (j=Y1;j<=ny;j++) for (i=X1-1;i<=nx;i++) {
    diag.ubt_flux[j][i] = Datu[j][i]*ubtav[j][i];
  }
  if (diag.vbt_flux != NULL) for (j=Y1-1;j<=ny;j++) for (i=X1;i<=nx;i++) {
    diag.vbt_flux[j][i] = Datv[j][i]*vbtav[j][i];
  }

/*  Offer various barotropic terms for averaging.                     */
  if (do_ave > 0) {
    if (av_ref_PFu_bt > 0) average_field((double (*)[NYMEM][NXMEM])
                             diag.PFu_bt, av_ref_PFu_bt);
    if (av_ref_PFv_bt > 0) average_field((double (*)[NYMEM][NXMEM])
                             diag.PFv_bt, av_ref_PFv_bt);
    if (av_ref_Coru_bt > 0) average_field((double (*)[NYMEM][NXMEM])
                              diag.Coru_bt, av_ref_Coru_bt);
    if (av_ref_Corv_bt > 0) average_field((double (*)[NYMEM][NXMEM])
                              diag.Corv_bt, av_ref_Corv_bt);
    if (av_ref_Nonlnu_bt > 0) average_field((double (*)[NYMEM][NXMEM])
                                diag.Nonlnu_bt, av_ref_Nonlnu_bt);
    if (av_ref_Nonlnv_bt > 0) average_field((double (*)[NYMEM][NXMEM])
                                diag.Nonlnv_bt, av_ref_Nonlnv_bt);
    if (av_ref_ubt_flux > 0) average_field((double (*)[NYMEM][NXMEM])
                               diag.ubt_flux, av_ref_ubt_flux);
    if (av_ref_vbt_flux > 0) average_field((double (*)[NYMEM][NXMEM])
                               diag.vbt_flux, av_ref_vbt_flux);
  }

}

void btcalc(double u[NZ][NYMEM][NXMEM], double v[NZ][NYMEM][NXMEM],
            double h[NZ][NYMEM][NXMEM], int lbt, int calculate_bt) {
/* btcalc calculates the barotropic velocities from the full velocity */
/* and thickness fields, determines the fraction of the total water   */
/* column in each layer at velocity points, and determines a correct- */
/* ive fictitious mass source that will drive the barotropic estimate */
/* of the free surface height toward the baroclinic estimate.         */

/* Arguments: u - Zonal velocity, in m s-1.                           */
/*  (in)      v - Meridional velocity, in m s-1.                      */
/*  (in)      h - Layer thickness, in m.                              */
/*  (in)      lbt - The time level of the barotropic velocities to    */
/*                  determine from the layer velocities.              */
/*  (in)      calculate_bt - A flag to indicate whether the barotropic*/
/*                           velocities and fractional thicknesses are*/
/*                           to be calculated.                        */

  double hatutot[NXMEM];    /* The sum of the layer thicknesses       */
  double hatvtot[NXMEM];    /* interpolated to the u & v grid points. */
  double Ihatutot[NXMEM];   /* Ihatutot and Ihatvtot are the inverses */
  double Ihatvtot[NXMEM];   /* of hatutot and hatvtot, both in m-1.   */
  int i, j, k;

/*    This section interpolates thicknesses onto u & v grid points    */
/*  with the second order accurate estimate h = 2*(h+ * h-)/(h+ + h-).*/
  if (calculate_bt) {
    for (j=Y1-1;j<=ny+1;j++) {
      for (i=X1-1;i<=nx+1;i++) {
        frhatu[0][j][i] = 2.0*(h[0][j][i+1] * h[0][j][i]) /
                                  (h[0][j][i+1] + h[0][j][i]);
        frhatv[0][j][i] = 2.0*(h[0][j+1][i] * h[0][j][i]) /
                                  (h[0][j+1][i] + h[0][j][i]);
        hatutot[i] = frhatu[0][j][i];
        hatvtot[i] = frhatv[0][j][i];
      }
      for (k=1;k<=NZ-1;k++) {
        for (i=X1-1;i<=nx+1;i++) {
          frhatu[k][j][i] = 2.0*(h[k][j][i+1] * h[k][j][i]) /
                                    (h[k][j][i+1] + h[k][j][i]);
          frhatv[k][j][i] = 2.0*(h[k][j+1][i] * h[k][j][i]) /
                                    (h[k][j+1][i] + h[k][j][i]);
          hatutot[i] += frhatu[k][j][i];
          hatvtot[i] += frhatv[k][j][i];
        }
      }
      for (i=X1-1;i<=nx+1;i++) {
        Ihatutot[i] = 1.0/hatutot[i]; Ihatvtot[i] = 1.0/hatvtot[i];
        ubt[lbt][j][i] = 0.0; vbt[lbt][j][i] = 0.0;
      }
      for (k=0;k<=NZ-1;k++) {
        for (i=X1-1;i<=nx+1;i++) {
          frhatu[k][j][i] *= Ihatutot[i];
          frhatv[k][j][i] *= Ihatvtot[i];
          ubt[lbt][j][i] += frhatu[k][j][i] * u[k][j][i];
          vbt[lbt][j][i] += frhatv[k][j][i] * v[k][j][i];
        }
      }
    }
  }

  for (j=Y1;j<=ny;j++) {
    double eta_h[NXMEM];    /* The free surface height determined from*/
                            /* the sum of the layer thicknesses, in m.*/
    for (i=X1;i<=nx;i++)
      eta_h[i] = h[0][j][i] - D[j][i];
    for (k=1;k<=NZ-1;k++) for (i=X1;i<=nx;i++) eta_h[i] += h[k][j][i];
    if (calculate_bt) for (i=X1;i<=nx;i++)
      eta_cor[j][i] = eta_h[i] - eta[lbt][j][i];
    else for (i=X1;i<=nx;i++)
      eta_cor[j][i] += eta_h[i] - eta[lbt][j][i];
  }
}

void barotropic_init(double u[NZ][NYMEM][NXMEM], double v[NZ][NYMEM][NXMEM],
                     double h[NZ][NYMEM][NXMEM], struct bt_vars_ptrs *bt) {
/*   barotropic_init initializes a number of time-invariant fields    */
/* usedin the barotropic calculation and initializes any barotropic   */
/* fields that have not already been initialized.                     */

/* Arguments: u - Zonal velocity, in m s-1.                           */
/*  (in)      v - Meridional velocity, in m s-1.                      */
/*  (in)      h - Layer thickness, in m.                              */
/*  (out)     bt - pointers (set here) to a collection of variables   */
/*                 related to the split time stepping scheme.         */
  double q[NYMEM][NXMEM];
  int i, j, k;

  /* Set the pointers to the variables in bt.                         */
  bt->ubtav = ubtav; bt->vbtav = vbtav; bt->IDatu = IDatu; bt->IDatv = IDatv;
  bt->gtot = gtot; bt->frhatu = frhatu; bt->frhatv = frhatv; bt->etaav = eta[2];

  if (!query_initialized(eta[0][0])) {
    for (j=Y1;j<=ny;j++) for (i=X1;i<=nx;i++) {
      eta[0][j][i] = -D[j][i];
      for (k=0;k<=NZ-1;k++)
        eta[0][j][i] += h[k][j][i];
    }
  }

  if ((!query_initialized(ubtav[0])) || (!query_initialized(vbtav[0]))) {
    btcalc(u,v,h,0,1);

    for (j=Y1-1;j<=ny+1;j++) for (i=X1-1;i<=nx+1;i++) {
      ubtav[j][i] = ubt[0][j][i];
      vbtav[j][i] = vbt[0][j][i];
    }
  }

/*   Calculate the constant coefficients for the Coriolis force       */
/* terms in the barotropic momentum equations.                        */

  for (j=Y1-1;j<=ny+1;j++) for (i=X1-1;i<=nx+1;i++) {
    q[j][i] = f[j][i] *
         ((DXDYh(j,i) + DXDYh(j+1,i+1)) + (DXDYh(j,i+1) + DXDYh(j+1,i))) /
         ((DXDYh(j,i) * D[j][i] + DXDYh(j+1,i+1) * D[j+1][i+1]) +
          (DXDYh(j,i+1) * D[j][i+1] + DXDYh(j+1,i) * D[j+1][i]));
  }

  for (j=Y1;j<=ny;j++) for (i=X1;i<=nx;i++) {
# ifndef SADOURNY
    amer[j][i-1] = (D[j][i] + D[j][i-1]) *
                   ((q[j][i] + q[j-1][i-1]) + q[j][i-1]) / 24.0;
    bmer[j][i] = (D[j][i+1] + D[j][i]) *
                 (q[j][i] + (q[j][i-1] + q[j-1][i])) / 24.0;
    cmer[j+1][i] = (D[j+1][i+1] + D[j+1][i]) *
                   (q[j][i] + (q[j][i-1] + q[j+1][i])) / 24.0;
    dmer[j+1][i-1] = (D[j+1][i] + D[j+1][i-1]) *
                     ((q[j][i] + q[j+1][i-1]) + q[j][i-1]) / 24.0;
# else
    amer[j][i-1] = (D[j][i] + D[j][i-1]) * q[j][i-1] / 8.0;
    bmer[j][i] = (D[j][i+1] + D[j][i]) * q[j][i] / 8.0;
    cmer[j+1][i] = (D[j+1][i+1] + D[j+1][i]) * q[j][i] / 8.0;
    dmer[j+1][i-1] = (D[j+1][i] + D[j+1][i-1]) * q[j][i-1] / 8.0;
# endif
  }

  for (j=Y1;j<=ny;j++) for (i=X1;i<=nx;i++) {
# ifndef SADOURNY
    azon[j][i] = (D[j][i+1] + D[j+1][i+1]) *
                 (q[j][i] + (q[j][i+1] + q[j-1][i])) / 24.0;
    bzon[j][i] = (D[j][i] + D[j+1][i]) *
                 (q[j][i] + (q[j][i-1] + q[j-1][i])) / 24.0;
    czon[j][i] = (D[j-1][i] + D[j][i]) *
                 ((q[j][i] + q[j-1][i-1]) + q[j-1][i]) / 24.0;
    dzon[j][i] = (D[j-1][i+1] + D[j][i+1]) *
                 ((q[j][i] + q[j-1][i+1]) + q[j-1][i]) / 24.0;
# else
    azon[j][i] = (D[j][i+1] + D[j+1][i+1]) * q[j][i] / 8.0;
    bzon[j][i] = (D[j][i] + D[j+1][i]) * q[j][i] / 8.0;
    czon[j][i] = (D[j-1][i] + D[j][i]) * q[j-1][i] / 8.0;
    dzon[j][i] = (D[j-1][i+1] + D[j][i+1]) * q[j-1][i] / 8.0;
# endif
  }

/*   Calculate other constants that are used for btstep.              */

  for (j=Y1;j<=ny;j++) for (i=X1-1;i<=nx;i++) {
    Datu[j][i] = umask[j][i] * (D[j][i+1] + D[j][i]) * 0.5 * DYu(j,i);
    IDatu[j][i] = umask[j][i] * 2.0 / (D[j][i+1] + D[j][i]);
  }
  for (j=Y1-1;j<=ny;j++) for (i=X1;i<=nx;i++) {
    Datv[j][i] = vmask[j][i] * (D[j+1][i] + D[j][i]) * 0.5 * DXv(j,i);
    IDatv[j][i] = vmask[j][i] * 2.0 / (D[j+1][i] + D[j][i]);
  }
# ifdef BOUND_BT_CORRECTION
  for (j=Y1;j<=ny;j++) for (i=X1;i<=nx;i++) {
    eta_cor_bound[j][i] = IDXDYh(j,i) * 0.1 * HIM_params.maxvel *
       ((Datu[j][i-1] + Datu[j][i]) + (Datv[j][i] + Datv[j-1][i]));
  }
# endif

  pass_var_2d(ubtav, 1, 1, 1, 0.0, 1, 0, TO_SOUTH+TO_EAST, U_PT+VECTOR);
  pass_var_2d(vbtav, 1, 1, 1, 0.0, 1, 0, TO_NORTH+TO_WEST, V_PT+VECTOR);

}

void register_barotropic_restarts(void) {
/* This subroutine is used to register any fields from barotropic.c   */
/* that should be written to or read from the restart file.           */
  vardesc vd[] = {
      {"sfc","Free surface Height",'h','1','s',"meter", 'd'},
      {"ubtav","Time mean barotropic zonal velocity",'u','1','s',
               "meter second-1", 'd'},
      {"vbtav","Time mean barotropic meridional velocity",'v','1','s',
               "meter second-1", 'd'}};

  register_restart_field(eta[0][0], eta[1][0], vd[0], 0);
  register_restart_field(ubtav[0], NULL, vd[1], 0);
  register_restart_field(vbtav[0], NULL, vd[2], 0);

}
#endif                                                       /* SPLIT */
