/**********************************************************************
 *                   GNU General Public License                       *
 * This file is a part of HIM.                                        *
 *                                                                    *
 * HIM is free software; you can redistribute it and/or modify it and *
 * are expected to follow the terms of the GNU General Public License *
 * as published by the Free Software Foundation; either version 2 of  *
 * the License, or (at your option) any later version.                *
 *                                                                    *
 * HIM is distributed in the hope that it will be useful, but WITHOUT *
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public   *
 * License for more details.                                          *
 *                                                                    *
 * For the full text of the GNU General Public License,               *
 * write to: Free Software Foundation, Inc.,                          *
 *           675 Mass Ave, Cambridge, MA 02139, USA.                  *
 * or see:   http://www.gnu.org/licenses/gpl.html                     *
 **********************************************************************/

/********+*********+*********+*********+*********+*********+*********+*
 *                                                                    *
 *  By Robert Hallberg, October 2001                                  *
 *                                                                    *
 *    This subroutine determines the strength of the barotropic body  *
 *  forcing for the DOME runs.  This forcing will cause the transport *
 *  through a specified control section of water in the densest       *
 *  layers to achieve a specified value averaged over a few days.     *
 *  The body forcing itself is applied in the back-channel through    *
 *  the barotropic momentum equations.                                *
 *                                                                    *
 *  Variables written all in capital letters are defined in init.h    *
 *                                                                    *
 *     A small fragment of the grid is shown below:                   *
 *                                                                    *
 *    j+1  x ^ x ^ x   At x:  q, f                                    *
 *    j+1  > o > o >   At ^:  v                                       *
 *    j    x ^ x ^ x   At >:  u, bt_body_force                        *
 *    j    > o > o >   At o:  h, D, buoy, tr, T, S, Rml, ustar        *
 *    j-1  x ^ x ^ x                                                  *
 *        i-1  i  i+1  At x & ^:                                      *
 *           i  i+1    At > & o:                                      *
 *                                                                    *
 *  The boundaries always run through q grid points (x).              *
 *                                                                    *
 ********+*********+*********+*********+*********+*********+*********+*/

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <init.h>
#include <HIM.h>
#include <metrics.h>

extern double D[NYMEM][NXMEM];       /* Basin depth, in m.            */

extern struct params HIM_params;   /* HIM_params is a structure that  */
                                   /* contains a number of parameters */
                                   /* of the simulation which can be  */
                                   /* modified from a namelist file   */
                                   /* at run time.                    */

extern double geolath[NYMEM][NXMEM]; /* The geographic latitude at h, */
extern double geolatq[NYMEM][NXMEM]; /* q, u, or v points in degrees  */
extern double geolatu[NYMEM][NXMEM]; /* of latitude or m.             */
extern double geolatv[NYMEM][NXMEM];
extern double geolonh[NYMEM][NXMEM]; /* The geographic longitude at h,*/
extern double geolonq[NYMEM][NXMEM]; /* q, u, or v points in degrees  */
extern double geolonu[NYMEM][NXMEM]; /* of longitude or m.            */
extern double geolonv[NYMEM][NXMEM];


extern int pe_here;                  /* The current processor's label.*/
extern int nx, ny;                   /* The largest index value in the*/
                                     /* x- and y- directions of the   */
                                     /* local computational domain.   */

double bt_body_force[NYMEM][NXMEM];  /* The pattern of the body force.*/
double bt_body_force_stren = 0.0;    /* The strength of the body force*/
                                     /* in units of m s-2.            */


void determine_body_force(double vh[NZ][NYMEM][NXMEM], double dt) {
  static int j_ave = -1;      /* The y-index of the control section.  */
  static int i_start_ave = NXMEM; /* The starting and ending x-indices*/
  static int i_end_ave = 0;   /* of the control section.              */
  int k_start = NZ/2;         /* This layer and denser are the ones   */
                              /* transport is summed.                 */
  double D_edge = 300.0;      /* The thickness of the dense plume at  */
                              /* the inner edge.                      */
  double lat_ave = 800.0;     /* The latitude of the control section. */
  double vh_sum;        /* The transport through the control section. */
  double vh_sum_target = -5.0e6;    /* The target transport in m3 s-1.*/
  static double vh_sum_ave = -3.0e6; /* The running mean transport.   */
  double lon_start = 950.0; /* The starting and stopping longitudes   */
  double lon_end = 1150.0;  /* of the control section.                */
  double I_run_ave = 1.0/(0.5*86400.0); /* The inverse e-folding time */
                                        /* for the running mean in s. */
  double I_Area;            /* The inverse of the cross-sectional     */
                            /* area of the dense plume, m-2.          */
  double I_time = 1.0/(2.0*86400.0); /* The damping rate for the      */
                                     /* forcing, in s-1.              */
  static int first_call = 1;
  static long call_no = 0;
  int i, j, k;

  I_Area = 1.0 / ( D_edge*sqrt(D_edge*(G_EARTH/RHO_0)*2.0)*1.0e4 );

  if (first_call == 1) {
    /*   Find the indices of the latitude and range of longitude upon */
    /* which to integrate the mass fluxes.                            */
    for (j=Y1;j<=ny;j++)
      if ((geolatu[j][X1] >= lat_ave) && (geolatu[j-1][X1] < lat_ave)) {
        j_ave = j; break;
      }
    if ((geolonu[Y1][X1-1] > lon_end) || (geolonu[Y1][nx] < lon_start)) {
      /* Do nothing. */
    } else {
      if (geolonu[Y1][X1-1] >= lon_start) i_start_ave = X1;
      else for (i=X1;i<=nx;i++)
        if ((geolonu[j][i] > lon_start) && (geolonu[j][i-1] <= lon_start)) {
          i_start_ave = i; break;
        }
      if (geolonu[Y1][nx] <= lon_end) i_end_ave = nx;
      else for (i=X1;i<=nx;i++)
        if ((geolonu[j][i] >= lon_end) && (geolonu[j][i-1] < lon_end)) {
          i_end_ave = i; break;
        }
    }

    /* This sets up the pattern for the forcing in the back channel.  */
    for (j=Y1;j<=ny;j++) for (i=X1;i<=nx;i++) {
      if ((D[j][i] > MINIMUM_DEPTH) && (geolatu[j][i] > 1000.0)) {
        if ((geolonu[j][i] >= 300.0) && (geolonu[j][i] <= 800.0))
          bt_body_force[j][i] = 1.0;
        else if ((geolonu[j][i] >= 200.0) && (geolonu[j][i] <= 300.0))
          bt_body_force[j][i] = (geolonu[j][i]-200.0)/100.0;
        else if ((geolonu[j][i] >= 800.0) && (geolonu[j][i] <= 900.0))
          bt_body_force[j][i] = (900.0-geolonu[j][i])/100.0;
        else bt_body_force[j][i] = 0.0;
      }
      else bt_body_force[j][i] = 0.0;
    }

    first_call = 0;
  }

  /* Integrate the meridional mass flux through the control section.  */
  vh_sum = 0.0;
  if (j_ave > 0) {
    for (k=k_start;k<NZ;k++) for (i=i_start_ave;i<=i_end_ave;i++) {
      vh_sum += vh[k][j_ave][i];
    }
  }
  sum_fields(&vh_sum,1);
  vh_sum_ave = vh_sum + exp(-dt*I_run_ave) * (vh_sum_ave - vh_sum);

  bt_body_force_stren -= 4.0*I_Area*dt*I_time*I_time*(vh_sum_target - vh_sum_ave);

  if ((pe_here == 0) && ((call_no % 48) == 0))
    printf("body force strength = %g - vh_sum = %g (~%g), target %g.\n",
        bt_body_force_stren,vh_sum,vh_sum_ave,vh_sum_target);

  call_no++;
}
